/*
 *  Unix SMB/CIFS implementation.
 *  OSF DCE ncacn_ip_tcp funnel
 *  Copyright (C) Luke Howard                       2002.
 *  Copyright (C) Andrew Tridgell              1992-1998,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1998,
 *  Copyright (C) Jeremy Allison                    1999.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 *
 * (c) Copyright 1989 OPEN SOFTWARE FOUNDATION, INC.
 * (c) Copyright 1989 HEWLETT-PACKARD COMPANY
 * (c) Copyright 1989 DIGITAL EQUIPMENT CORPORATION
 * To anyone who acknowledges that this file is provided "AS IS"
 * without any express or implied warranty:
 *                 permission to use, copy, modify, and distribute this
 * file for any purpose is hereby granted without fee, provided that
 * the above copyright notices and this notice appears in all source
 * code copies, and that none of the names of Open Software
 * Foundation, Inc., Hewlett-Packard Company, or Digital Equipment
 * Corporation be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  Neither Open Software Foundation, Inc., Hewlett-
 * Packard Company, nor Digital Equipment Corporation makes any
 * representations about the suitability of this software for any
 * purpose.
 *
 */

/*
 * For each service exported over ncacn_np, look it up in the
 * endpoint mapper and contact it on the local host using
 * ncacn_ip_tcp. Presently there is no authentication.
 */

#ifdef NCACN_IP_TCP_FUNNEL

#ifndef PRIVATE_DIR /* XXX */
#define PRIVATE_DIR "/usr/xad/private"
#endif

#include <includes.h>

#include <dce/ep.h>

typedef struct ncacn_ip_tcp_funnel {
	struct pipes_struct sp;
	BOOL ep_resolved;
	uint16 ep_port;
	struct in_addr ep_addr;
	int sd;
	pstring skey_file;
} ncacn_ip_tcp_funnel;

/*
 * XXX memory leak -- freeing RPC runtime allocated memory
 * appears to crash smbd. Probably a matter of using the
 * correct free() function, but even rpc_ss_client_free()
 * does not work. For now, leak memory.
 */
#define RPC_FREE(x)

/* #define RPC_FREE(x) SAFE_FREE(x) */

#define MAX_ENTS 7
#define PIPE            "\\PIPE\\"
#define PIPELEN         (sizeof(PIPE) - 1)

extern struct pipe_id_info pipe_names[];

/* These macros are a quick way to negotiate the endpoint maps. */

#define CHECK_FLOOR_SPACE(ptr, tower, len) do { \
	if (ptr - tower->tower_octet_string >= tower->tower_length - len) \
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: out of floor space\n")); \
		RPC_FREE(tower); \
		continue; \
	} while (0)

#define CLIMB_HALF_FLOOR(ptr, tower) do { \
	uint16 floor_len; \
	CHECK_FLOOR_SPACE(ptr, tower, 2); \
	floor_len = SVAL(ptr, 0); \
	ptr += 2; \
	CHECK_FLOOR_SPACE(ptr, tower, floor_len); \
	ptr += floor_len; \
	} while (0)

#define CLIMB_FLOOR(ptr, tower) do { \
	CLIMB_HALF_FLOOR(ptr, tower); \
	CLIMB_HALF_FLOOR(ptr, tower); \
	} while (0)

/**
 * Open up a connection to the ncacn_ip_tcp server
 * that we are proxying to.
 */
static BOOL ncacn_ip_tcp_open_funnel_socket(ncacn_ip_tcp_funnel *p)
{
	extern pstring user_socket_options;

	if (p == NULL)
		return False;

	if (p->ep_resolved == False)
		return False;

	p->sd = open_socket_out(SOCK_STREAM, &p->ep_addr, p->ep_port, LONG_CONNECT_TIMEOUT);
	if (p->sd < 0) {
		DEBUG(0, ("ncacn_ip_tcp_open_funnel_socket: error connecting to server\n"));
		return False;
	}

	set_socket_options(p->sd, user_socket_options);

	return True;
}

/**
 * Close the funnel to the ncacn_ip_tcp server.
 */
static void ncacn_ip_tcp_close_funnel_socket(ncacn_ip_tcp_funnel *p)
{
	if (p == NULL)
		return;

	if (p->sd > -1) {
		if (close(p->sd) < 0) {
			DEBUG(0, ("ncacn_ip_tcp_close_funnel_socket: error closing socket; ignoring\n"));
		}
	} else {
		DEBUG(0, ("ncacn_ip_tcp_close_funnel_socket: socket was not open\n"));
	}
}

/**
 * Resolve a pipe name in the endpoint mapper, given
 * its UUID and version.
 */
static BOOL ncacn_ip_tcp_resolve_endpoint(ncacn_ip_tcp_funnel *p)
{
	RPC_IFACE *iface;
	struct pipe_id_info *infop;
	rpc_binding_handle_t ep_handle;
	rpc_protseq_vector_t *protseq_vector;
	unsigned32 st;
	int i;
	fstring protseq_str;
	ept_entry_t entries[MAX_ENTS];
	unsigned32 num_ents;
	rpc_if_id_t if_id;
	ept_lookup_handle_t entry_handle = NULL;
	unsigned_char_p_t sUuid;

	if (p == NULL) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: called with NULL pipe\n"));
		return False;
	}

	if (p->ep_resolved) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: endpoint %s already resolved\n", p->sp.name));
		return True;
	}

	/* Find the UUID and version for the RPC interface. */
	for (iface = NULL, infop = &pipe_names[0];
	     infop->client_pipe != NULL; infop++) {
		uuid_to_string((uuid_t *)&infop->abstr_syntax.uuid, &sUuid, &st);

		DEBUG(9, ("ncacn_ip_tcp_resolve_endpoint: checking client_pipe %s server_pipe %s UUID %s\n", infop->client_pipe, infop->server_pipe, (st == rpc_s_ok) ? sUuid : "<unknown>"));
		if (st == rpc_s_ok) {
			rpc_string_free(&sUuid, &st);
		}

		if (strncmp(infop->client_pipe, PIPE, PIPELEN) == 0 &&
		    strcasecmp(infop->client_pipe + PIPELEN, p->sp.name) == 0) {
			iface = &infop->abstr_syntax;
			break;
		}
	}

	if (iface == NULL) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: could not find abstract syntax for pipe %s\n", p->sp.name));
		return False;
	}

	rpc_network_inq_protseqs(&protseq_vector, &st);
	if (st != rpc_s_ok) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: rpc_network_inq_protseqs failed\n"));
		return False;
	}

	for (i = 0; i < protseq_vector->count; i++) {
		int slen = strlen(protseq_vector->protseq[i]);
		if ((slen + 2) > sizeof(protseq_str))
			continue;
		safe_strcpy(protseq_str, protseq_vector->protseq[i], slen);
		safe_strcpy(&protseq_str[slen], ":", 1);
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: attempting to resolve %s\n", protseq_str));
		rpc_binding_from_string_binding(protseq_str, &ep_handle, &st);
		if (st == rpc_s_ok) {
			break;
		}
	}

	if (i >= protseq_vector->count) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: rpc_binding_from_string_handle failed\n"));
		rpc_protseq_vector_free(&protseq_vector, &st);
		return False;
	}

	rpc_protseq_vector_free(&protseq_vector, &st);
	memcpy(&if_id.uuid, &iface->uuid, sizeof(iface->uuid));
	if_id.vers_minor = iface->version >> 16;
	if_id.vers_major = iface->version & 0xffff;

	uuid_to_string(&if_id.uuid, &sUuid, &st);
	DEBUG(3, ("ncacn_ip_tcp_resolve_endpoint: UUID %s major=%d minor=%d\n", (st == rpc_s_ok) ? sUuid : "<unknown>", if_id.vers_major, if_id.vers_minor));
	if (st == rpc_s_ok) {
		rpc_string_free(&sUuid, &st);
	}

	/* Now, contact the endpoint mapper. */
	(*ept_v3_0_c_epv.ept_lookup)(ep_handle,
		rpc_c_ep_match_by_if,       /* inquiry_type */
		NULL,                       /* object */
		&if_id,                     /* interface */
		rpc_c_vers_compatible,      /* vers_option */
		&entry_handle,              /* entry_handle */
		MAX_ENTS,                   /* max_ents */
		&num_ents,                  /* num_ents */
		entries,                    /* entries */
		&st);
	if (st != rpc_s_ok) {
		DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: ept_lookup failed status=%08x\n", st));
		rpc_binding_free(&ep_handle, &st);
		return False;
	}

	DEBUG(5, ("ncacn_ip_tcp_resolve_endpoint: ept_lookup returned %d entries\n", num_ents));

	/*
	 * Now, we get an RPC tower back from the EP mapper.
	 * Unfortunately, the runtime is not very helpful in
	 * decoding it, unless you have access to the internal
	 * headers (which would make this even more of an
	 * abstraction violation). So, we need to take apart
	 * the tower octet string bit by bit. This parsing
	 * is based on looking at traces in Ethereal: each
 	 * tower has a number of floors, with a LHS and a
	 * RHS. We skip the first two floors as they contain
 	 * the interface and syntax UUIDs. The remaining
	 * floors should contain { cn, tcp, ip } in that
 	 * order. The lower two floors also contain the
 	 * TCP port and IP address in the RHS, respectively.
	 * Note that we are fairly fussy about length coding
	 * being compact (ie. no trailing zero bytes). We
	 * should fix this...
	 */
	for (i = 0; i < num_ents; i++) {
		twr_p_t tower = entries[i].tower;
		unsigned_char_p_t ptr = tower->tower_octet_string;
		uint16 num_floors, len;

		DEBUG(5, ("ncacn_ip_tcp_resolve_endpoint: found %s\n", entries[i].annotation));

		dump_data(5, tower->tower_octet_string, tower->tower_length);

		CHECK_FLOOR_SPACE(ptr, tower, 2);

		/* Get the floor count. */
		num_floors = SVAL(ptr, 0);
		if (num_floors != 5) {
			DEBUG(5, ("ncacn_ip_tcp_resolve_endpoint: got %d floors, expected 5\n", num_floors));
			continue;
		}
		ptr += 2;

		/* Climb the first UUID floor. */
		CLIMB_FLOOR(ptr, tower);
		/* Climb the second UUID floor. */
		CLIMB_FLOOR(ptr, tower);

		/* Now, we should be on the third, which is connection oriented. */
		CHECK_FLOOR_SPACE(ptr, tower, 2);
		len = SVAL(ptr, 0);
		ptr += 2;

		if (len != 1) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: third floor protocol value (CN) too long\n"));
			RPC_FREE(tower);
			continue;
		}

		if (*((uint8 *)ptr) != 0x0b) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: cannot parse third floor protocol; got %02x instead of 0x0b\n", *((uint8 *)ptr) & 0xff));
			RPC_FREE(tower);
			continue;
		}
		ptr += len;

		/* skip RHS */
		CLIMB_HALF_FLOOR(ptr, tower);

		/* Now, we should be on the fourth floor, which is TCP. */
		CHECK_FLOOR_SPACE(ptr, tower, 2);
		len = SVAL(ptr, 0);
		ptr += 2;

		if (len != 1) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: fourth floor protocol value (TCP) too long\n"));
			RPC_FREE(tower);
			continue;
		}

		if (*((uint8 *)ptr) != 0x07) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: cannot parse fourth floor protocol; got %02x instead of 0x07\n", *((uint8 *)ptr) & 0xff));
			RPC_FREE(tower);
			continue;
		}
		ptr += len;

		/* RHS of floor should have port */
		CHECK_FLOOR_SPACE(ptr, tower, 2);
		len = SVAL(ptr, 0);
		ptr += 2;

		/* Get port. */
		CHECK_FLOOR_SPACE(ptr, tower, len);
		p->ep_port = ntohs(SVAL(ptr, 0));
		ptr += len;

		/* LHS of floor 5 should have IP constant. */
		CHECK_FLOOR_SPACE(ptr, tower, 2);
		len = SVAL(ptr, 0);
		ptr += 2;

		if (len != 1) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: fifth floor protocol value (IP) too long\n"));
			RPC_FREE(tower);
			continue;
		}

		if (*((uint8 *)ptr) != 0x09) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: cannot parse fourth floor protocol; got %02x instead of 0x09\n", *((uint8 *)ptr) & 0xff));
			RPC_FREE(tower);
			continue;
		}
		ptr += len;
	
		/* RHS of floor 5 should have IP address */
		CHECK_FLOOR_SPACE(ptr, tower, 2);
		len = SVAL(ptr, 0);
		ptr += 2;

		if (len != 4) {
			DEBUG(0, ("ncacn_ip_tcp_resolve_endpoint: cannot parse fifth floor IP address length %d\n", len));
			RPC_FREE(tower);
			continue;
		}

		/* Huh, network byte order? */
		p->ep_addr.s_addr = IVAL(ptr, 0);
		ptr += len;
		p->ep_resolved = True;

		RPC_FREE(tower);

		DEBUG(4, ("Resolved pipe %s to ncacn_ip_tcp:%s[%d]\n", p->sp.name, inet_ntoa(p->ep_addr), p->ep_port));

		break;
	}

	(*ept_v3_0_c_epv.ept_lookup_handle_free)(ep_handle, &entry_handle, &st);

	rpc_binding_free(&ep_handle, &st);

	return p->ep_resolved;
}

/**
 * Create a new funnel to a ncacn_ip_tcp RPC
 * server.
 */
void *ncacn_ip_tcp_make_funnel(char *pipe_name,
	connection_struct *conn,
	uint16 vuid)
{
	ncacn_ip_tcp_funnel *p;
	user_struct *vuser = get_valid_user_struct(vuid);
	static BOOL runtime_init = False;

	DEBUG(4, ("ncacn_ip_tcp_make_funnel: %s (vuid %d)\n", pipe_name, vuid));

	if (vuser == NULL && vuser != UID_FIELD_INVALID) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: vuid %d did not map to a valid vuser struct\n", vuid));
	}

	p = (ncacn_ip_tcp_funnel *)malloc(sizeof(*p));
	if (p == NULL) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: out of memory\n"));
		return NULL;
	}

	ZERO_STRUCTP(p);
	p->sd = -1;

	p->sp.mem_ctx = talloc_init();
	if (p->sp.mem_ctx == NULL) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: talloc_init failed\n"));
		SAFE_FREE(p);
		return NULL;
	}

	if (!init_pipe_handle_list(&p->sp, pipe_name)) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: init_pipe_handle_list failed\n"));
		talloc_destroy(p->sp.mem_ctx);
		SAFE_FREE(p);
		return NULL;
	}

	if (!prs_init(&p->sp.in_data.data, MAX_PDU_FRAG_LEN, p->sp.mem_ctx, MARSHALL)) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: prs_init failed\n"));
		SAFE_FREE(p);
		return NULL;
	}

	fstrcpy(p->sp.name, pipe_name);

	if (ncacn_ip_tcp_resolve_endpoint(p) == False) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: could not resolve endpoint %s\n", pipe_name));
		SAFE_FREE(p);
		return NULL;
	}

	if (ncacn_ip_tcp_open_funnel_socket(p) == False) {
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: could not open socket for funnel pipe %s\n", pipe_name));
		SAFE_FREE(p);
		return NULL;
	}

	p->sp.conn = conn;
	p->sp.conn->num_files_open++;
	p->sp.vuid = vuid;

	p->sp.ntlmssp_chal_flags = 0;
	p->sp.ntlmssp_auth_validated = False;
	p->sp.ntlmssp_auth_requested = False;

	p->sp.pipe_bound = False;
	p->sp.fault_state = False;
	p->sp.endian = RPC_LITTLE_ENDIAN;

	ZERO_STRUCT(p->sp.pipe_user);

	p->sp.pipe_user.uid = -1;
	p->sp.pipe_user.gid = -1;

	p->skey_file[0] = '\0';

	if (vuser != NULL) {
		int fd;
		struct sockaddr_in name;
		size_t namelen = sizeof(name);

		memcpy(p->sp.session_key, vuser->session_key, sizeof(vuser->session_key));
		p->sp.pipe_user.nt_user_token = dup_nt_token(vuser->nt_user_token);

		if (getsockname(p->sd, &name, &namelen) < 0) {
			DEBUG(0, ("ncacn_ip_tcp_make_funnel: warning - could not get local socket name\n"));
			ncacn_ip_tcp_close_funnel(p);
			return NULL;
		}

		snprintf(p->skey_file, sizeof(p->skey_file), "%s/.skey_%s[%d]", PRIVATE_DIR, inet_ntoa(name.sin_addr), ntohs(name.sin_port));
		DEBUG(0, ("ncacn_ip_tcp_make_funnel: storing session key in file %s\n", p->skey_file));

		fd = sys_creat(p->skey_file, S_IREAD|S_IWRITE);
		if (fd < 0) {
			DEBUG(0, ("ncacn_ip_tcp_make_funnel: sys_creat(%s, S_IREAD|S_IWRITE) failed: %s\n", p->skey_file, strerror(errno)));
			ncacn_ip_tcp_close_funnel(p);
			return NULL;
		}
		if (fchown(fd, vuser->uid, vuser->gid) < 0) {
			DEBUG(0, ("ncacn_ip_tcp_make_funnel: fchown(%d, %d, %d) failed\n", fd, vuser->uid, vuser->gid));
			ncacn_ip_tcp_close_funnel(0);
			return NULL;
		}
		if (sys_write(fd, p->sp.session_key, sizeof(p->sp.session_key)) != sizeof(p->sp.session_key)) {
			DEBUG(0, ("ncacn_ip_tcp_make_funnel: sys_write failed while writing session key\n"));
			ncacn_ip_tcp_close_funnel(p);
			return NULL;
		}
		if (close(fd) < 0) {
			DEBUG(0, ("ncacn_ip_tcp_make_funnel: close session key file failed\n"));
		}
	}

	p->sp.in_data.pdu_needed_len = 0;
	p->sp.in_data.pdu_received_len = 0;

	p->sp.out_data.current_pdu_len = 0;
	p->sp.out_data.current_pdu_sent = 0;
	p->sp.out_data.data_sent_length = 0;

	prs_init(&p->sp.out_data.rdata, 0, p->sp.mem_ctx, MARSHALL);

	DEBUG(4, ("Created ncacn_ip_tcp funnel pipe %s\n", pipe_name));

	return p;
}

/**
 * Read shim. Not sure whether we need to be clever
 * with respect to splitting and reassembling PDUs
 * or whether we can just funnel them across. Testing
 * will reveal all.
 */
ssize_t ncacn_ip_tcp_read_from_funnel(void *np_conn,
	char *data,
	size_t n,
	BOOL *is_data_outstanding)
{
	ncacn_ip_tcp_funnel *p = (ncacn_ip_tcp_funnel *)np_conn;
	ssize_t ret, selrtn;
	fd_set fds;
	struct timeval timeout;

	*is_data_outstanding = False;

	if (p->sd < 0) {
	   return -1;
	}

	/* Check whether n > MAX_PDU_FRAG_LEN??? */
	ret = sys_read(p->sd, data, n);

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	FD_ZERO(&fds);
	FD_SET(p->sd, &fds);

	selrtn = sys_select(p->sd + 1, &fds, NULL, NULL, &timeout);
	if (selrtn > 0 && FD_ISSET(p->sd, &fds)) {
		*is_data_outstanding = True;
	}

	DEBUG(0, ("ncacn_ip_tcp_read_from_funnel: read %d bytes from funnel (maximum %d); %smore data to read\n", ret, n, *is_data_outstanding ? "" : "no "));

	return ret;
}

/**
 * Write shim. Not sure whether we need to be clever
 * with respect to splitting and reassembling PDUs
 * or whether we can just funnel them across. Testing
 * will reveal all.
 */
ssize_t ncacn_ip_tcp_write_to_funnel(void *np_conn,
	char *data,
	size_t n)
{
	ncacn_ip_tcp_funnel *p = (ncacn_ip_tcp_funnel *)np_conn;
	ssize_t ret;

	if (p->sd < 0) {
		return -1;
	}

	ret = write_socket(p->sd, data, n);

	DEBUG(0, ("ncacn_ip_tcp_write_to_funnel: wrote %d of %d bytes to funnel\n", ret, n));

	return ret;
}

/**
 * Close the RPC pipe handle.
 */
BOOL ncacn_ip_tcp_close_funnel(void *np_conn)
{
	ncacn_ip_tcp_funnel *p = (ncacn_ip_tcp_funnel *)np_conn;
	if (p == NULL) {
		DEBUG(0, ("ncacn_ip_tcp_close_funnel: invalid pipe\n"));
		return False;
	}
	prs_mem_free(&p->sp.out_data.rdata);
	prs_mem_free(&p->sp.in_data.data);

	if (p->sp.mem_ctx != NULL) {
		talloc_destroy(p->sp.mem_ctx);
	}

	delete_nt_token(&p->sp.pipe_user.nt_user_token);

	SAFE_FREE(p->sp.pipe_user.groups);
	p->sp.conn->num_files_open--;

	ncacn_ip_tcp_close_funnel_socket(p);

	if (p->skey_file[0]) {
		unlink(p->skey_file);
	}

	ZERO_STRUCT(p);
	SAFE_FREE(p);

	DEBUG(0, ("ncacn_ip_tcp_close_funnel: closed funnel\n"));

	return True;
}

#endif /* NCACN_IP_TCP_FUNNEL */

